#!/usr/bin/perl
use strict;

my $tagCharacters = "a-zA-z0-9_";
my $tagName = "\/{0,1}[$tagCharacters]*";
my(%tagValidAll,%tagValid,%tagRequired,%tagMultiple,$tagBase,@params);
my $verbosity = "";
my $passed = 0;

if (!@ARGV) {
   print q{
Usage: xmvalidate [-skipped] [-checked] [-details] model [document ...]

Without flags the program verifies the model and then validates all
listed documents, showing any errors.

The flags add additional information to the output, as follows.

-checked: list all documents checked.
-skipped: list any documents skipped because the model does not apply.
-details: show full details for each checked document.

The flags can actually appear anywhere in the parameter list. The
first non-flag paramater is the name of the model file, and all the
others are the names (or filespecs) of files to check against it.

Version 0.2, 6 September 2005.

};
   exit;
};

for (@ARGV) {
   if ($_ eq "-checked" || $_ eq "-skipped" || $_ eq "-details") {
      $verbosity .= $_;
   } else {
      push(@params,$_);
   };
};

my($returnvalue,$r) = xmValidate(@params);
print $r;

sub xmValidate{
   my(@thisFiles) = @_;
   my ($returncode,$returnvalue);
   my $r = "";
   my $thisXM = iFileRead($thisFiles[0]);

   $r .= "\n   Checking model $thisFiles[0].\n\n";
   if ($thisXM) {
      $thisXM =~ s|[ \t\n\r]||g;
      ($returncode,$returnvalue) = xmCheckSyntaxModel($thisXM);
      $r .= $returnvalue;
      return(0,$r) if (!$returncode);
   } else {
      $r .= "** No model.\n";
      return(0,$r);
   };

   for (@thisFiles[1 .. $#thisFiles]) {
      ($returncode,$returnvalue) = xmCheckSyntax(iFileRead($_),$_);
      $passed += $returncode;
      $r .= $returnvalue;
   };

   $r .= "\n\n   " . (scalar(@thisFiles) - 1) . " examined, $passed valid according to this model.\n\n";
   return(1,$r);
};


sub xmCheckSyntaxModel{
   my($thisXML) = @_;
   my(@thisTags,@tagStack,$errorUnlessNext,$thisTagName,$thisTagCode,$thisStrayText,$key,$r);
   (@thisTags) = $thisXML =~ /(\<$tagName\>[.?+*]{0,1})/sg;
   ($tagBase,undef) = $thisTags[0] =~ /\<(.*?)\>([.?+*]{0,1})/;
   for (@thisTags) {
      if (/^\<\//) {
         if ($_ eq "</$tagStack[-1]>") {
            pop(@tagStack);
         } else {
            $r .= "** Unmatched tag <$tagStack[-1]>\n\n";
            return(0,$r);
         };
      } else {
         ($thisTagName) = /\<(.*?)\>/;
         push(@tagStack,$thisTagName);
      };
   };
   if (@tagStack) {
      $r .= "** Unclosed tag <$tagStack[-1]>\n\n";
      return(0,$r);
   } else {
      $r .= "   Tag stack OK.\n" if ($verbosity =~ /details/);
   };
   for (@thisTags) {
      if ($errorUnlessNext && ($_ ne "</$errorUnlessNext>")) {
         $r .= "** Duplicate definition for container tag <$errorUnlessNext>.\n\n";
         return(0,$r);
      } else {
         $errorUnlessNext = "";
      };
      if (/^\<\//) {
         pop(@tagStack);
      } else {
         ($thisTagName,$thisTagCode) = /\<(.*?)\>([.?+*]{0,1})/;
         if (!$thisTagCode) {
            $r .= "** Missing .?+* quantifier for tag <$thisTagName>.\n\n";
            return(0,$r);
         };
         if ($tagValid{$thisTagName}) {
            $r .= "** Duplicate definition for container tag <$errorUnlessNext>.\n\n";
            return(0,$r);
         } elsif ($tagValidAll{$thisTagName}) {
            $errorUnlessNext = $thisTagName;
         };
         $tagValidAll{$thisTagName} = "valid";
         $tagValid{$tagStack[-1]} or $tagValid{$tagStack[-1]} .= "valid: ";
         $tagValid{$tagStack[-1]} .= "<$thisTagName>";
         if ($thisTagCode eq "+" || $thisTagCode eq ".") {
            $tagRequired{$tagStack[-1]} or $tagRequired{$tagStack[-1]} .= "required: ";
            $tagRequired{$tagStack[-1]} .= "<$thisTagName>";
         };
         if ($thisTagCode eq "+" || $thisTagCode eq "*") {
            $tagMultiple{$tagStack[-1]} or $tagMultiple{$tagStack[-1]} .= "multiple: ";
            $tagMultiple{$tagStack[-1]} .= "<$thisTagName>";
         };
         push(@tagStack,$thisTagName);
      };
   };

   $thisStrayText = $thisXML;
   $thisStrayText =~ s|\<\?.*?\?\>||sg;
   for $key (keys %tagValidAll) {
      if ($tagValid{$key}) {
         $thisStrayText =~ s|\<\/{0,1}$key\>[.?+*]{0,1}||sg;
      } else {
         $thisStrayText =~ s|\<$key\>.\<\/$key\>||sg;
      };
   };
   if ($thisStrayText) {
      $r .= "** Stray text in element context.\n\n";
      $r .= "$thisStrayText\n\n";
      return(0,$r);
   };

   $r .= "      $tagBase (base)\n";
   $r .= "         " . $tagValid{$tagBase} . "\n";
   $r .= "         " . $tagRequired{$tagBase} . "\n" if ($tagRequired{$tagBase});
   $r .= "         " . $tagMultiple{$tagBase} . "\n" if ($tagMultiple{$tagBase});
   for $key (sort keys %tagValid) {
      if ($key && $key ne $tagBase) {
         $r .= "      $key\n";
         $r .= "         " . $tagValid{$key} . "\n";
         $r .= "         " . $tagRequired{$key} . "\n" if ($tagRequired{$key});
         $r .= "         " . $tagMultiple{$key} . "\n" if ($tagMultiple{$key});
      };
   };

   $r .= "\n\n" if ($verbosity);
   return(1,$r);
};


sub xmCheckSyntax{
   my($thisXML,$thisFile) = @_;
   my(@thisBaseXMLs,$thisBaseXML,@thisAllTags,@thisCheckTags,@tagStack,%countHash,@countHashes,$thisTagName,$thisStrayText,$key,$r);
   my $thisErrors = "";

   (@thisBaseXMLs) = $thisXML =~ /(\<$tagBase\>.*?\<\/$tagBase\>)/sg;
   if (!@thisBaseXMLs) {
      $r .= "\n" if ($verbosity =~ /(checked)|(details)/);
      $r .= "-- No <$tagBase> base in $thisFile.\n" if ($verbosity =~ /skipped/);
      $r .= "\n\n" if ($verbosity =~ /(checked)|(details)/);
      return(0,$r);
   };
   $r .= "   Checking document $thisFile (" . length($thisXML) . ")" if ($verbosity =~ /(checked)|(details)/);
   $r .= "\n\n" if ($verbosity =~ /details/);
   for $thisBaseXML (@thisBaseXMLs) {
      for $key (keys %tagValidAll) {
         $thisBaseXML =~ s|\<$key\>\<\/$key\>||sg unless $tagValid{$key};
      };
      (@thisAllTags) = $thisBaseXML =~ /\<($tagName)\>/sg;
      for (@thisAllTags) {
         $thisTagName = $_;
         $thisTagName =~ s|^\/||;
         push(@thisCheckTags,$_) if ($tagValidAll{$thisTagName});
      };
      for (@thisCheckTags) {
         if (/^\//) {
            if ($_ eq ("/" . $tagStack[-1])) {
               my @thisTagsRequired = $tagRequired{$tagStack[-1]} =~ /\<(.*?)\>/g;
               for my $thisTagRequired (@thisTagsRequired) {
                  $thisErrors .= "\n++ " . $tagStack[-1] . " missing required <$thisTagRequired> in $thisFile.\n" if ($countHash{$thisTagRequired} < 1);
               };
               for $key (keys %countHash) {
                  $thisErrors .= "\n++ " . $tagStack[-1] . " contains " . $countHash{$key} . " uses of non-multiple tag <$key> in $thisFile.\n" if ($countHash{$key} > 1 && !($tagMultiple{$tagStack[-1]} =~ /$key/) );
               };
               pop(@tagStack);
               %countHash = %{pop(@countHashes)};
            } else {
               $r .= "\n** Unmatched tag <$tagStack[-1]> in $thisFile.\n\n";
               return(0,$r);
            };
         } else {
            if ($tagValid{$tagStack[-1]} =~ /\<$_\>/) {
               $r .= sprintf("%" . (scalar(@tagStack) + 2) * 3 . "s"," ") . "$_\n" if ($verbosity =~ /details/);
               $countHash{$_}++;
               push(@tagStack,$_);
               my %pushCountHash = %countHash;
               push(@countHashes,\%pushCountHash);
               undef %countHash;
            } else {
               $r .= "\n** Tag $_ not allowed within $tagStack[-1] in $thisFile.\n\n";
               return(0,$r);
            };
         };
      };
      if (@tagStack) {
         $r .= "\n** Unclosed tag <$tagStack[-1]> in $thisFile.\n\n";
         return(0,$r);
      };

      $thisStrayText = $thisBaseXML;
      for $key (keys %tagValidAll) {
         if ($tagValid{$key}) {
            $thisStrayText =~ s|\<\/{0,1}$key\>[.?+*]{0,1}||sg;
         } else {
            $thisStrayText =~ s|\<$key\>.*?\<\/$key\>||sg;
         };
      };
      $thisStrayText =~ s|^[ \t\n\r]*||sg;
      $thisStrayText =~ s|[ \t\n\r]*$||sg;
      if ($thisStrayText) {
         $r .= "\n** Stray text in element context in $thisFile.\n\n";
         $r .= "$thisStrayText\n\n";
         return(0,$r);
      };
   };

   if ($thisErrors) {
      $r .= "$thisErrors\n";
      return(0,$r);
   } else {
      $r .= "\n      Valid!\n\n\n" if ($verbosity =~ /details/);
      $r .= " - valid.\n" if ($verbosity =~ /checked/ && !($verbosity =~ /details/));
      return(1,$r);
   };
};



sub iFileRead{
   my ($thisFileName) = @_;
   my $thisFileContent = "";
   local $/;
   open(FILE,$thisFileName);
   $thisFileContent = <FILE>;
   close(FILE);
   return $thisFileContent;
};